-----------------------------------------------------------------------------------------------
-- Client Lua Script for Tardis
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
require "GameLib"
require "math"

-----------------------------------------------------------------------------------------------
-- TarDis Module Definition
-----------------------------------------------------------------------------------------------
local TarDis = {} 

local kiFonts = Apollo.GetGameFonts() 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------

local tTarDisProperties = {
--	["Fonts"] 			= { "CRB_Interface16_BBO",	"CRB_Interface16_BBO" },
--	["tColors"] 		= { { 1, 1, 1, 1 }, { 0, 1, 0, 1  } },
--	["tBGColors"] 		= { { 0.5, 0.5, 0.5, 0.5 }, { 0.5, 0.5, 0.5, 0.5  } },
--	["daDefaultColor"]	= { 1.0, 0.45, 0.45, 1 },
--	["tDistances"] 		= { 20, 200 },
	bMoveFrame 		= true,
	tFrame			= 	
		{ 
			name="TargetDistance", anchors = { 0.5, 0.5, 0.5, 0.5 }, 
			offsets = { -65, -20,  65, 20 }, movable  = true, show = 1,
		},
	tRows = { 
		{
			sFont = "CRB_Interface16_BBO",
			nFontIndex = 79,
			tColor = { 1, 1, 1, 1 },
			tBGColor = { 0.5, 0.5, 0.5, 0.5 },
			sDistance = 20
		},
		{
			sFont = "CRB_Interface16_BBO",
			nFontIndex = 79,
			tColor = { 0, 1, 0, 1 },
			tBGColor = { 0.5, 0.5, 0.5, 0.5 },
			sDistance = 200
		},
		{
			sFont = "CRB_Interface16_BBO",
			nFontIndex = 79,
			tColor = { 1, 1, 1, 1 },
			tBGColor = { 0.5, 0.5, 0.5, 0.5 },
			sDistance = "Default"
		}
	},
	nConfig = 0.2
}

	
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function TarDis:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here
	o.tItems = {} -- keep track of all the list items
	o.wndFontSlider = nil -- fontslider
	o.maxDist = 200 * 200 -- maxdist to check
	
	o.tProps = {} -- For saving and loading properties
	o.tOriginalProps = {}
	o.tColors = {}
	o.tBGColors = {}
	o.daDefaultColor = nil
	o.nSelectedIndex = nil
	
	o.sFont = "CRB_Interface16_BBO"
	o.nColIndex = -1 -- Index for the current distance
	o.sAdjust = ""
	
	
	
	return o
end

function TarDis:Init()
	local bHasConfigureFunction = true
	local strConfigureButtonText = "Target Distance"
	local tDependencies = {
		-- "UnitOrPackageName",
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 

-----------------------------------------------------------------------------------------------
-- TarDis OnLoad
-----------------------------------------------------------------------------------------------
function TarDis:OnLoad()
	-- If config file doesn't exist restore isn't called. We set original tables and colors up
	self.tProps = TableUtil:Copy( tTarDisProperties ) 
	self:setColors( self.tProps.tRows )
--	self:setDefaultColor( self.tProps[ "daDefaultColor" ] )
    -- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("TarDis.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
	
end



-----------------------------------------------------------------------------------------------
-- TarDis OnDocLoaded
-----------------------------------------------------------------------------------------------
function TarDis:OnDocLoaded()
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
	    self.wndMain = Apollo.LoadForm(self.xmlDoc, "TarDisForm", nil, self)
	    self.wndOff = Apollo.LoadForm(self.xmlDoc, "Form2", nil, self)

	
		if self.wndMain == nil or self.wndOff == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end
			
		-- item list
		self.wndGrid = self.wndMain:FindChild("Grid")
		self.wndItemList = self.wndMain:FindChild("TarDisItems")

	    self.wndMain:Show(false, true)
	    self.wndOff:Show(true, true)

		Apollo.RegisterSlashCommand("tardis", "OnTarDisOn", self)
		Apollo.RegisterEventHandler("VarChange_FrameCount", "OnFrame", self)				
		
		self:setFrame( self.tProps["tFrame"] )
		
		self.wndMain:FindChild("DefaultColorBox"):SetBGColor( self.daDefaultColor )
	end
end

-- on Configuration button
function TarDis:OnConfigure()
	self:OnTarDisOn()
end

-----------------------------------------------------------------------------------------------
-- TarDis OnSave and OnRestore
-----------------------------------------------------------------------------------------------



function TarDis:OnRestore(eType, tSavedData)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then
		return
	end
	self.tProps = TableUtil:Copy( tTarDisProperties )

	for property, tData in pairs(tTarDisProperties ) do
		if tSavedData[property] ~= nil then
			if tData and type( tData ) == type( tSavedData[property] ) then
				self.tProps[property] = tSavedData[property]
			else
				self.tProps[property] = tTarDisProperties [property]
			end
		else
			self.tProps[property] = tTarDisProperties [property]
		end
		
		
		if property == "tRows" then
			self:setColors( self.tProps.tRows )
--			self:setColors( self.tProps.tRows, self.tColors )

--		elseif property == "daDefaultColor" then
--			self:setDefaultColor( self.tProps.daDefaultColor )			
		end
	end	
	
	if tSavedData[ "nConfig" ] == nil then
		self.tProps["tRows"] = {}
		for index = 1, #tSavedData.tDistances do
			
			local row = {} 
			
			row.sDistance = tSavedData.tDistances[index]
			row.tColor = tSavedData.tColors[index]
			row.tBGColor = { 0.5, 0.5, 0.5, 0.5 }
			row.sFont = "CRB_Interface16_BBO"
			
			self.tProps.tRows[index] = row
		end
		local defaultRow = {}
		row.sDistance = ""
		row.tColor = tSavedData.tDefaultColor
		row.tBGColor = { 0.5, 0.5, 0.5, 0.5 }
		row.sFont = "CRB_Interface16_BBO"
				
		table.insert( self.tProps.tRows, row )		
		self:setColors( self.tProps.tRows )
	end
		
end

function TarDis:setColors( props )
	for colIndx, tData in pairs( props ) do
		local tColData = tData.tColor
		self.tColors[ colIndx ] =  CColor.new( tColData[1], tColData[2], tColData[3], tColData[4] )
		tColData = tData.tBGColor
		self.tBGColors[ colIndx ] =  CColor.new( tColData[1], tColData[2], tColData[3], tColData[4] )
	end
end

--[[
function TarDis:setColors( props, index )
	for colIndx, tColData in pairs( props ) do
		self.tColors[ colIndx ] =  CColor.new( tColData[1], tColData[2], tColData[3], tColData[4] )
	end
end
]]

function TarDis:setDefaultColor( color )
	self.daDefaultColor = CColor.new( color[1], color[2], color[3], color[4] )
end

function TarDis:OnSave(eType)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then
		return
	end
	
	self:updateFrame()
	
	local tSave = {}
	for property,tData in pairs(tTarDisProperties ) do
		tSave[property] = self.tProps[property]
	end
	
	return tSave
end


function TarDis:setFrame( props )
	local wnd  = self.wndOff

	wnd:SetAnchorPoints( unpack( props.anchors ) ) 
	wnd:SetAnchorOffsets( unpack( props.offsets ) )
end

function TarDis:updateFrame( )
	local tFrame = self.tProps["tFrame"]
	local wnd  = self.wndOff

	tFrame.anchors = { wnd:GetAnchorPoints() } -- Convert multible return values to table
	tFrame.offsets  = { wnd:GetAnchorOffsets() } -- Convert multible return values to table


end

function TarDis:updateItems() -- Updates the values of options
--	self.tProps["tDistances"] = {} 	-- Resetting tables if adding/deleting occured
--	self.tProps["tColors"] = {} 	-- does this leak memory?


--	local tDists = self.tProps["tDistances"]
--	local tColors = self.tProps["tColors"]
	
	local tRows = self.tRows

	for idx,wnd in ipairs(self.tItems) do
		local nDist = tonumber( wnd:FindChild("EditBox"):GetText() )
				
		local colorBox = wnd:FindChild("ColorBox")
		local col = colorBox:GetBGColor()
	
		local sortIndx = 1
		local i, tmp = next( tRows )
		
		-- Add objects while keeping array sorted
		while tmp ~= nil and i ~= nil and nDist ~= "Default" and tmp.sDistance < nDist do
			sortIndx = sortIndx + 1
			i, tmp = next( tDists, i )
		end
		
		table.insert( tColors, sortIndx, { col.r, col.g, col.b, col.a } )
		table.insert( tDists, sortIndx, nDist )

	end
	
	
	
		
	
	local col  = self.wndMain:FindChild( "DefaultColorBox" ):GetBGColor()
	self.tProps[ "daDefaultColor" ] = { col.r, col.g, col.b, col.a }
end


-----------------------------------------------------------------------------------------------
-- TarDis Functions
-----------------------------------------------------------------------------------------------
-- Define general functions here

-- on SlashCommand "/tardis"
function TarDis:OnTarDisOn()




	self.tOriginalProps = TableUtil:Copy( self.tProps ) 

	self.wndOff:Invoke()
	self.wndMain:Invoke() -- show the window, configuration window on top
	
	-- populate the item list
	self:PopulateItemList()


end


-- Called during frame -------------------------
function TarDis:OnFrame()
	local wnd = self.wndOff
	
	-- Do your timer-related stuff here.
	if wnd  == nil then return end
	
	local target = GameLib.GetTargetUnit()
	local player  = GameLib.GetPlayerUnit()

	if player == nil or target == nil then 
		local nIndx = #self.tProps.tRows
		
		local sFont = self.tProps.tRows[nIndx].sFont
		if self.sFont ~= sFont then
			self.sFont = sFont
			wnd:SetFont( sFont )
		end
		
		wnd:SetTextColor( self.tColors[nIndx] )
		wnd:SetBGColor( self.tBGColors[nIndx] )

		self.wndOff:SetText("No Target")
		return 
	end
	
	
	--local strName = drPlayer and drPlayer:GetName() or "player"
	local xdist = ( target:GetPosition().x - player:GetPosition().x ) ^ 2
	local ydist = ( target:GetPosition().y - player:GetPosition().y ) ^ 2
	local zdist = ( target:GetPosition().z - player:GetPosition().z ) ^ 2
	
	local fDist = math.sqrt( xdist + ydist + zdist )
	local tRows = self.tProps.tRows
	
	local nIndx, tRow = next( tRows, nil )

	while( nIndx ~= nil and tRow ~=nil and tRow.sDistance ~= "Default" and fDist > tRow.sDistance ) do
		nIndx, tRow = next( tRows, nIndx )		
	end
	
	--if nDistCompare == nil or nIndx == nil then
	--	self.wndOff:SetTextColor( self.daDefaultColor )

	--else
	wnd:SetTextColor( self.tColors[nIndx] )
	wnd:SetBGColor( self.tBGColors[nIndx] )
	--end
	
	wnd:SetText( string.format("%.2f", fDist ) .. "m" )

	local sFont = tRow.sFont
	if self.sFont ~= sFont then
		self.sFont = sFont
		wnd:SetFont( sFont )
	end
end


-----------------------------------------------------------------------------------------------
-- TarDisForm Functions
-----------------------------------------------------------------------------------------------
-- when the OK button is clicked
function TarDis:OnOK()

	-- Sorting items, slow sorting:
	local rows = self.tProps.tRows
	
	local count = #rows 
	for i = 1, count do
		if rows[i].sDistance == "Default" then
			rows[i], rows[count] = rows[count], rows[i]
		else
			for j = i, count do
				if rows[j].sDistance ~= "Default" and rows[i].sDistance > rows[j].sDistance then
					rows[i], rows[j] = rows[j], rows[i]
				end
			end
		end
	end
	--self:updateItems() -- First update items in list
	
	-- then reset all colors, because of reordering items
	--self.tColors = {}
	
	-- setup backup colors
	self:setColors( self.tProps.tRows )
	--self:setDefaultColor( self.tProps[ "daDefaultColor" ] )
	
	self.wndMain:Close() -- hide the window
end

-- when the Cancel button is clicked
function TarDis:OnCancel()
	self.tProps =  TableUtil:Copy(  self.tOriginalProps )	
	-- remove colors
	--self.tColors = {}
	
	-- setup backup colors
	self:setColors( self.tProps.tRows )
	--self:setDefaultColor( self.tProps[ "daDefaultColor" ] )
	
	self.wndMain:Close() -- hide the window
end


function TarDis:OnAddItemButtonPressed( wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, 
	nLastRelativeMouseY, bDoubleClick, bStopPropagation )
	
	-- Check that we are using right handler and pressing left mouse button.. 
	-- This function is never called without left click though
	if wndHandler ~= wndControl or eMouseButton ~= GameLib.CodeEnumInputMouse.Left   then
		return
	end
	
	-- Add new item
	local nDist = 100
	local color = CColor.new( 1, 1, 1, 1 )
	local bgColor = CColor.new( 0.5, 0.5, 0.5, 0.5 )
	
	local newIndx = #self.tColors + 1
	
	local row = {
		sFont = "CRB_Interface16_BBO",
		nFontIndex = 79,
		tColor = { 1, 1, 1, 1 },
		tBGColor = { 0.5, 0.5, 0.5, 0.5 },
		sDistance = 200
	}
	
	self.tProps.tRows[ newIndx ] = row
	self.tColors[ newIndx ] = color
	self.tBGColors[ newIndx ] = bgColor 

	self:AddItem( newIndx, row )
	
	-- now all the item are added, call ArrangeChildrenVert to list out the list items vertically
	self.wndItemList:ArrangeChildrenVert()
	 
end

function TarDis:OnRemoveItemButtonPressed( wndHandler, wndControl, eMouseButton )

	-- Check that we are using right handler and pressing left mouse button.. 
	-- This function is never called without left click though
	if wndHandler ~= wndControl or eMouseButton ~= GameLib.CodeEnumInputMouse.Left  then
		return
	end
	
	
	if self.nSelectedIndex == nil then
		return
	end
	
	-- Cant delete default
	if self.nSelectedIndex == #self.tProps.tRows then
		return
	end
	
	-- First unselect item
	local oldIndx = self.nSelectedIndex

	self:ListItemUnselect()

	-- Removes item from different places	
	self.tItems[ oldIndx ]:Destroy()	
	table.remove( self.tItems,  oldIndx )
	table.remove( self.tColors, oldIndx  )
	table.remove( self.tProps.tRows, oldIndx )
	for i = oldIndx, #self.tColors do
		local daItem = self.tItems[ i ]
		
		local dataIndx = daItem:GetData()
		dataIndx.indx = i

		daItem:SetData( dataIndx )
		
		for ind, val in pairs( daItem:GetChildren() ) do
			val:SetData( dataIndx  )		
		end
	end
	
	-- now all the item are added, call ArrangeChildrenVert to list out the list items vertically
	self.wndItemList:ArrangeChildrenVert()

end

function TarDis:OnCellSelect( wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY )
	if wndHandler ~= wndControl or eMouseButton ~= GameLib.CodeEnumInputMouse.Left then return end
	
--	self.wndGrid
end

-----------------------------------------------------------------------------------------------
-- ItemList Functions
-----------------------------------------------------------------------------------------------

-- populate item list
function TarDis:PopulateItemList()
	-- make sure the item list is empty to start with
	self:DestroyItemList()
	
    -- add #of tProps items
--	local index = 1
	for i, row in pairs( self.tProps.tRows )	do
        self:AddItem( i, row )  -- self.tProps["tDistances"][ i ], self.tColors[ i ], self.tProps.tFonts[i] )
		i = i + 1
	end
	
	-- now all the item are added, call ArrangeChildrenVert to list out the list items vertically
	self.wndItemList:ArrangeChildrenVert()
	
	self.nSelectedIndex = nil
end

-- clear the item list
function TarDis:DestroyItemList()
	-- destroy all the wnd inside the list
	for idx,wnd in ipairs(self.tItems) do
		wnd:Destroy()
	end

	-- clear the list item array
	self.tItems = {}
	self.nSelectedIndex = nil
end

-- add an item into the item list
function TarDis:AddItem( i, row ) -- nDist, color, sFont )
	local nDist = row.sDistance
	local color = self.tColors[ i ]
	local bgColor = self.tBGColors[ i ]
	local sFont = row.sFont

	-- load the window item for the list item
	local wnd = Apollo.LoadForm(self.xmlDoc, "ListItem", self.wndItemList, self)
	
	local editWindow = wnd:FindChild("EditWindow")
	local editBox = wnd:FindChild("EditBox")
	local colorBox = wnd:FindChild("ColorBox")
	local bgColBox = wnd:FindChild("BGColor")
	--local comboBox = wnd:FindChild("FontComboBox")
	local fontSelector = wnd:FindChild("FontSelector")
	
	if nDist == "Default" then
		editWindow :SetText( nDist ) 
		editWindow :Enable( true )
		wnd:EnsureChildVisible( editWindow )
		
		
		editBox:SetText( nDist )
		editBox:Enable( false )	
	else
		editBox:SetText( nDist ) 
		editBox:Enable( true )
		wnd:EnsureChildVisible( editBox )

		editWindow :Enable( false )
	end
	
	editBox:SetText( nDist ) 
	
	colorBox:SetBGColor( bgColor )
	colorBox:SetTextColor( color )

	bgColBox:SetBGColor( bgColor )
	bgColBox:SetTextColor( color )
	
	
--	for index, value in pairs( kiFonts ) do
--		comboBox:AddItem( value.name )		
--	end
	
	--comboBox:SetText( sFont )
	fontSelector:SetText( sFont )	
	bgColBox:SetFont( sFont )
	colorBox:SetFont( sFont )

	-- keep track of the window item created
	self.tItems[i] = wnd

	local dataIndx = {}
	dataIndx.indx = i

	local indexer = 
	
	wnd:SetData( dataIndx )
	self:AddDatas( wnd, dataIndx )
	--editBox:SetData( dataIndx )
	--bgColBox:SetData( dataIndx  )
	--colorBox:SetData( dataIndx  )
	--editWindow:SetData( dataIndx )
--	comboBox :SetData( dataIndx )
	--fontSelector :SetData( dataIndx )
end

function TarDis:AddDatas( item, data )
	local childs = item:GetChildren()
	for i,v in pairs( childs ) do
		v:SetData( data )
		self:AddDatas( v, data )
	end
end

-- when a list item is selected
function TarDis:OnListItemSelected(wndHandler, wndControl)
--	if wndHandler ~= wndControl then return end 
	self:ListItemUnselect() 
			
	self.nSelectedIndex = wndHandler:GetData().indx
	self.tItems[ self.nSelectedIndex ]:SetBGColor( "white" )
	
end

function TarDis:ListItemUnselect( )
	for i,v in pairs( self.tItems ) do
		v:SetBGColor( "black" )
	end
	self.nSelectedIndex = nil
end

---------------------------------------------------------------------------------------------------
-- ListItem Functions
---------------------------------------------------------------------------------------------------


function TarDis:OnEditChange( wndHandler, wndControl, strText )
	self:OnListItemSelected(wndHandler, wndControl)

	local num = tonumber( strText )
	if num == nil then				
		strText =  0
		wndHandler:SetText( strText ) 
	end	
	

	self.tProps.tRows[ wndHandler:GetData().indx ].sDistance = num
	
end


function TarDis:OnColorButtonPressed( wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, 
	nLastRelativeMouseY, bDoubleClick, bStopPropagation )
	
	-- To make sure the control is supposed to be same as handler. 
	-- Should always be the case since nothing should be on top of the box.
	if wndHandler ~= wndControl then
		return
	end
	
	if not ( wndHandler:GetName() == "BGColor" or wndHandler:GetName() == "ColorBox" ) then return end
	
	
	self.nSelectedIndex = wndHandler:GetData().indx
	local oldInd = self.nSelectedIndex

	
	local col 
	if wndControl:GetName() == "BGColor" then
		col = self.tBGColors[ self.nSelectedIndex ]
		self.sAdjust = "BGCol"
		
	else
		col = self.tColors[ self.nSelectedIndex ]
		self.sAdjust = "FontCol"
	end
	
	

	-- Color, hasAlpha, callback func, callback param
	ColorPicker.AdjustCColor( col, true, self.OnColorChanged, self  ) 
end

function TarDis:OnColorChanged( )

	if self.sAdjust == "BGCol" then
		local col = self.tBGColors[ self.nSelectedIndex ]
		self.tItems[ self.nSelectedIndex ]:FindChild("BGColor"):SetBGColor( col )	
		self.tItems[ self.nSelectedIndex ]:FindChild("ColorBox"):SetBGColor( col )	
		
		self.tProps.tRows[ self.nSelectedIndex ].tBGColor = { col.r, col.g, col.b, col.a }

	else

		local col = self.tColors[ self.nSelectedIndex ]

		self.tItems[ self.nSelectedIndex ]:FindChild("BGColor"):SetTextColor( col )	
		self.tItems[ self.nSelectedIndex ]:FindChild("ColorBox"):SetTextColor( col )

		self.tProps.tRows[ self.nSelectedIndex ].tColor = { col.r, col.g, col.b, col.a }
	end
	
end

function TarDis:OnFontBoxChanged( wndHandler, wndControl )
	if wndHandler ~= wndControl then return end
	local sFont = wndHandler:GetSelectedText()
	local nIndx = wndHandler:GetData().indx
	Print( sFont )
	self.tProps.tRows[ nIndx ].sFont = sFont
	
	self.tItems[ nIndx  ]:FindChild("BGColor"):SetFont( sFont )	
	self.tItems[ nIndx  ]:FindChild("ColorBox"):SetFont( sFont )

	
	
end


function TarDis:OnFontButton( wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation )
	if wndControl ~= wndHandler or eMouseButton ~= GameLib.CodeEnumInputMouse.Left then return end
	
	local nIndx =  wndHandler:GetData().indx
	local item = self.tProps.tRows[ nIndx ]
	local index = item.nFontIndex
	if index == nil then 
		index = 1 
		item.nFontIndex = 1
	end
	
	if wndControl:GetName() == "Left" then
		if index > 1 then 
			item.nFontIndex = index - 1
		end
	else
		if item.nFontIndex < #kiFonts then 
			item.nFontIndex = index + 1
		end
	end
	local sFont = kiFonts[ index ].name

	item.sFont = sFont
	self.tItems[ nIndx  ]:FindChild("BGColor"):SetFont( sFont )	
	self.tItems[ nIndx  ]:FindChild("ColorBox"):SetFont( sFont )
	wndHandler:GetParent():SetText( sFont )
end

-----------------------------------------------------------------------------------------------
-- TarDis Instance
-----------------------------------------------------------------------------------------------
local TarDisInst = TarDis:new()
TarDisInst:Init()
